@@ -7,11 +7,14 @@ from xml.etree.ElementTree import Element, SubElement, Comment, tostring
 from enum import Enum
 import asyncio
 import logging
+from datetime import datetime, timedelta
 
-# import config
 import aiohttp
 
 HAYWARD_API_URL = "https://www.haywardomnilogic.com/HAAPI/HomeAutomation/API.ashx"
+HAYWARD_AUTH_URL = "https://services-gamma.haywardcloud.net/auth-service/v2/login"
+HAYWARD_REFRESH_URL = "https://services-gamma.haywardcloud.net/auth-service/v2/refresh"
+HAYWARD_APP_ID = "tzwqg83jvkyurxblidnepmachs"
 
 _LOGGER = logging.getLogger("omnilogic")
 
@@ -23,6 +26,8 @@ class OmniLogic:
         self.systemname = None
         self.userid = None
         self.token = None
+        self.refresh_token = None
+        self.token_expiry = None
         self.verbose = True
         self.logged_in = False
         self.retry = 5
@@ -78,6 +83,10 @@ class OmniLogic:
         """
         Generic method to call API.
         """
+        # Check if authentication is needed
+        if self.token and self.token_expiry and datetime.now() >= self.token_expiry:
+            await self.authenticate()
+            
         payload = self.buildRequest(methodName, params)
 
         headers = {
@@ -124,36 +133,114 @@ class OmniLogic:
         return response
 
     async def _get_token(self):
-
-        params = {"UserName": self.username, "Password": self.password}
-
-        response = await self.call_api("Login", params)
-
-        if "There is no information" in response:
-            raise OmniLogicException("Failure getting token.")
-
-        else:
-            root = ElementTree.fromstring(response)
-            userid = root[1][2].text
-            token = root[1][3].text
-            # await self.close()
-
-            return {"token": token, "userid": userid}
+        """ Get a new authentication token using the new auth endpoint """
+        headers = {
+            "Content-Type": "application/json",
+            "X-HAYWARD-APP-ID": HAYWARD_APP_ID
+        }
+        
+        payload = {
+            "email": self.username,
+            "password": self.password
+        }
+        
+        _LOGGER.debug(f"Authenticating with URL: {HAYWARD_AUTH_URL}")
+        _LOGGER.debug(f"Using headers: {headers}")
+        _LOGGER.debug(f"Using payload structure: {list(payload.keys())}")
+        
+        try:
+            async with self._session.post(
+                HAYWARD_AUTH_URL, json=payload, headers=headers
+            ) as resp:
+                if resp.status != 200:
+                    error_text = await resp.text()
+                    _LOGGER.error(f"Authentication failed with status {resp.status}")
+                    _LOGGER.error(f"Error details: {error_text}")
+                    _LOGGER.error(f"Response headers: {resp.headers}")
+                    _LOGGER.error(f"Request URL: {HAYWARD_AUTH_URL}")
+                    _LOGGER.error(f"Request payload keys: {list(payload.keys())}")
+                    _LOGGER.error(f"Using username/email: {self.username[:3]}...{self.username[-3:] if len(self.username) > 6 else ''}")
+                    raise LoginException(f"Failed login: {resp.status}. Details: {error_text}")
+                
+                response = await resp.json()
+                _LOGGER.debug(f"Authentication successful. Response keys: {list(response.keys())}")
+                
+                # Set token expiry to 24 hours from now (refresh daily)
+                self.token_expiry = datetime.now() + timedelta(hours=24)
+                
+                return {
+                    "token": response.get("access_token"),
+                    "refresh_token": response.get("refresh_token"),
+                    "userid": response.get("user_id")
+                }
+                
+        except aiohttp.ClientConnectorError as e:
+            raise LoginException(f"Connection error: {e}")
 
     async def _get_new_token(self):
         return await self._get_token()
+        
+    async def _refresh_token(self):
+        """ Refresh the authentication token """
+        if not self.refresh_token:
+            # If no refresh token is available, get a new token instead
+            return await self._get_token()
+            
+        headers = {
+            "Content-Type": "application/json",
+            "X-HAYWARD-APP-ID": HAYWARD_APP_ID
+        }
+        
+        payload = {
+            "refresh_token": self.refresh_token
+        }
+        
+        try:
+            async with self._session.post(
+                HAYWARD_REFRESH_URL, json=payload, headers=headers
+            ) as resp:
+                if resp.status != 200:
+                    _LOGGER.warning("Token refresh failed, getting new token")
+                    # If refresh fails, fall back to getting a new token
+                    return await self._get_token()
+                
+                response = await resp.json()
+                
+                # Set token expiry to 24 hours from now
+                self.token_expiry = datetime.now() + timedelta(hours=24)
+                
+                return {
+                    "token": response.get("access_token"),
+                    "refresh_token": response.get("refresh_token"),
+                    "userid": self.userid  # Keep the existing user ID
+                }
+                
+        except aiohttp.ClientConnectorError as e:
+            _LOGGER.error(f"Token refresh connection error: {e}")
+            # If refresh fails with connection error, fall back to getting a new token
+            return await self._get_token()
 
     async def authenticate(self):
-
-        if not self.token:
+        """ Authenticate or refresh token if needed """
+        # Check if token needs refresh
+        if self.token and self.token_expiry and datetime.now() < self.token_expiry:
+            # Token is still valid, no action needed
+            return
+            
+        # Get new token or refresh existing token
+        if not self.token or not self.refresh_token:
             response = await self._get_new_token()
+        else:
+            response = await self._refresh_token()
 
-            if response != '{"Error":"Failed"}':
-                self.token = response["token"]
-                self.userid = response["userid"]
-            else:
-                self.token = ""
-                self.userid = ""
+        if response and "token" in response:
+            self.token = response["token"]
+            self.refresh_token = response.get("refresh_token")
+            self.userid = response["userid"]
+        else:
+            self.token = None
+            self.refresh_token = None
+            self.userid = None
 
     async def connect(self):
         """
@@ -292,6 +379,8 @@ class OmniLogic:
                             for light in BOW["ColorLogic-Light"]:
                                 if "V2-Active" not in light:
                                     light["V2-Active"] = "no"
+                                else:
+                                    light["V2-Active"] = "yes"
                                 bow_lights.append(light)
                         except:
                             if isinstance(
@@ -305,6 +394,8 @@ class OmniLogic:
                                 ]:
                                     if "V2-Active" not in light:
                                         light["V2-Active"] = "no"
+                                    else:
+                                        light["V2-Active"] = "yes"
                                     bow_lights.append(light)
                             else:
                                 light = configitem["Backyard"]["Body-of-water"][
@@ -312,6 +403,8 @@ class OmniLogic:
                                 ]
                                 if "V2-Active" not in light:
                                     light["V2-Active"] = "no"
+                                else:
+                                    light["V2-Active"] = "yes"
                                 bow_lights.append(light)
 
                     BOW = json.loads(BOW)
@@ -361,11 +454,15 @@ class OmniLogic:
                                         this_light = BOW["ColorLogic-Light"]
                                         if "V2-Active" not in this_light:
                                             this_light["V2-Active"] = "no"
+                                        else:
+                                            this_light["V2-Active"] = "yes"
                                         bow_lights.append(this_light)
                                         break
                                     else:
                                         if "V2-Active" not in light:
                                             light["V2-Active"] = "no"
+                                        else:
+                                            light["V2-Active"] = "yes"
                                         bow_lights.append(light)
                             except:
                                 bow_lights.append(BOW["ColorLogic-Light"])
@@ -692,7 +789,11 @@ class OmniLogic:
         return success
 
     def alarms_to_json(self, alarms):
-        alarmsXML = ElementTree.fromstring(alarms)
+        try:
+            alarmsXML = ElementTree.fromstring(alarms)
+        except:
+            raise OmniLogicException("Error loading Hayward data.")
+            
         alarmslist = []
 
         for child in alarmsXML:
@@ -716,7 +817,11 @@ class OmniLogic:
         return alarmslist
 
     def telemetry_to_json(self, telemetry, config_data, site_alarms):
-        telemetryXML = ElementTree.fromstring(telemetry)
+        try:
+            telemetryXML = ElementTree.fromstring(telemetry)
+        except:
+            raise OmniLogicException("Error loading Hayward data.")
+
         backyard = {}
 
         BOW = {}
@@ -758,6 +863,7 @@ class OmniLogic:
                     BOW_list = []
                     bow_lights = []
                     bow_relays = []
+                    bow_pumps = []
                     bow_heaters = []
                     relays = []
                     BOWname = ""
@@ -782,6 +888,7 @@ class OmniLogic:
                     BOW = {}
                     bow_lights = []
                     bow_relays = []
+                    bow_pumps = []
                     bow_heaters = []
 
                     BOWname = "BOW" + str(child.attrib["systemId"])
@@ -963,76 +1070,136 @@ class OmniLogic:
 
     async def get_telemetry_data(self):
         if self.token is None:
+            _LOGGER.debug("Token is None, attempting to connect")
             await self.connect()
         if len(self.systems) == 0:
+            _LOGGER.debug("No systems found, retrieving site list")
             await self.get_site_list()
 
         # assert self.token != "", "No login token"
         telem_list = []
 
         if self.token != "" and len(self.systems) != 0:
-            config_data = await self.get_msp_config_file()
-            
-            """
-            f = open("mspconfig_" + self.username + ".txt", "w")
-            f.write(str(config_data))
-            f.close()
-            """
-            
-            for system in self.systems:
-                # Get the right instance of the ID for this system
-                config_item = {}
-
-                for sys_data in config_data:
-
-                    if sys_data["MspSystemID"] == system["MspSystemID"]:
-                        config_item = sys_data
-
-                params = {"Token": self.token, "MspSystemID": system["MspSystemID"]}
-
-                telem = await self.call_api("GetTelemetryData", params)
+            try:
+                _LOGGER.debug(f"Getting MSP config file for {len(self.systems)} systems")
+                config_data = await self.get_msp_config_file()
+                _LOGGER.debug(f"Successfully retrieved MSP config data")
                 
-                params = {
-                    "Token": self.token,
-                    "MspSystemID": system["MspSystemID"],
-                    "Version": "0",
-                }
-
-                this_alarm = await self.call_api("GetAlarmList", params)
-
-                site_alarms = self.alarms_to_json(this_alarm)
-
-                if site_alarms[0].get("BowID") == "False":
-                    site_alarms = []
+                """
+                f = open("mspconfig_" + self.username + ".txt", "w")
+                f.write(str(config_data))
+                f.close()
+                """
                 
-                site_telem = self.telemetry_to_json(telem, config_item, self.alarms_to_json(this_alarm))
+                for system in self.systems:
+                    try:
+                        # Get the right instance of the ID for this system
+                        config_item = {}
+                        _LOGGER.debug(f"Processing system: {system['MspSystemID']} - {system.get('BackyardName', 'Unknown')}")
 
-                site_telem["BackyardName"] = config_item["BackyardName"]
-                site_telem["Msp-Vsp-Speed-Format"] = config_item["System"][
-                    "Msp-Vsp-Speed-Format"
-                ]
-                site_telem["Msp-Time-Format"] = config_item["System"]["Msp-Time-Format"]
-                site_telem["Units"] = config_item["System"]["Units"]
-                site_telem["Msp-Chlor-Display"] = config_item["System"][
-                    "Msp-Chlor-Display"
-                ]
-                site_telem["Msp-Language"] = config_item["System"]["Msp-Language"]
-                site_telem["Unit-of-Measurement"] = config_item["System"]["Units"]
-                site_telem["Alarms"] = site_alarms
+                        for sys_data in config_data:
+                            if sys_data["MspSystemID"] == system["MspSystemID"]:
+                                config_item = sys_data
 
-                if type(config_item["Backyard"]["Sensor"]) == dict:
-                  site_telem["Unit-of-Temperature"] = config_item["Backyard"]["Sensor"][
-                      "Units"
-                  ]
-                else:
-                  for sensor in config_item["Backyard"]["Sensor"]:
-                    if sensor["Name"] == "AirSensor":
-                      site_telem["Unit-of-Temperature"] = sensor["Units"]
+                        if not config_item:
+                            _LOGGER.warning(f"Could not find config data for system {system['MspSystemID']}")
+                            continue
 
-                telem_list.append(site_telem)
+                        params = {"Token": self.token, "MspSystemID": system["MspSystemID"]}
+                        _LOGGER.debug(f"Getting telemetry data for system {system['MspSystemID']}")
+
+                        telem = await self.call_api("GetTelemetryData", params)
+                        _LOGGER.debug(f"Successfully retrieved telemetry data for system {system['MspSystemID']}")
+                        
+                        params = {
+                            "Token": self.token,
+                            "MspSystemID": system["MspSystemID"],
+                            "Version": "0",
+                        }
+                        _LOGGER.debug(f"Getting alarm list for system {system['MspSystemID']}")
+
+                        this_alarm = await self.call_api("GetAlarmList", params)
+                        _LOGGER.debug(f"Successfully retrieved alarm list for system {system['MspSystemID']}")
+
+                        site_alarms = self.alarms_to_json(this_alarm)
+                        _LOGGER.debug(f"Processed alarms: {len(site_alarms)} found")
+
+                        if site_alarms[0].get("BowID") == "False":
+                            site_alarms = []
+                        
+                        _LOGGER.debug(f"Converting telemetry to JSON for system {system['MspSystemID']}")
+                        site_telem = self.telemetry_to_json(telem, config_item, self.alarms_to_json(this_alarm))
+                        _LOGGER.debug(f"Successfully converted telemetry to JSON for system {system['MspSystemID']}")
+
+                        site_telem["BackyardName"] = config_item["BackyardName"]
+                        
+                        try:
+                            site_telem["Msp-Vsp-Speed-Format"] = config_item["System"]["Msp-Vsp-Speed-Format"]
+                            site_telem["Msp-Time-Format"] = config_item["System"]["Msp-Time-Format"]
+                            site_telem["Units"] = config_item["System"]["Units"]
+                            site_telem["Msp-Chlor-Display"] = config_item["System"]["Msp-Chlor-Display"]
+                            site_telem["Msp-Language"] = config_item["System"]["Msp-Language"]
+                            site_telem["Unit-of-Measurement"] = config_item["System"]["Units"]
+                            site_telem["Alarms"] = site_alarms
+                        except KeyError as e:
+                            _LOGGER.error(f"Missing key in system config: {e}")
+                            _LOGGER.debug(f"Available system keys: {list(config_item.get('System', {}).keys())}")
+
+                        try:
+                            if "Sensor" in config_item["Backyard"]:
+                                sensors = config_item["Backyard"]["Sensor"]
+                                _LOGGER.debug("Found sensors in Backyard")
+                            else:
+                                if "Sensor" in config_item["Backyard"].get("Body-of-water", {}):
+                                    sensors = config_item["Backyard"]["Body-of-water"]["Sensor"]
+                                    _LOGGER.debug("Found sensors in Body-of-water")
+                                else:
+                                    sensors = {}
+                                    _LOGGER.debug("No sensors found")
+
+                            hasAirSensor = False
+
+                            if type(sensors) == dict and sensors != {}:
+                                site_telem["Unit-of-Temperature"] = sensors.get("Units","UNITS_FAHRENHEIT")
+
+                                if sensors["Name"] == "AirSensor":
+                                    hasAirSensor = True
+                                    _LOGGER.debug("Found AirSensor")
+                            else:
+                                for sensor in sensors:
+                                    if sensor["Name"] == "AirSensor":
+                                        site_telem["Unit-of-Temperature"] = sensor.get("Units","UNITS_FAHRENHEIT")
+                                        hasAirSensor = True
+                                        _LOGGER.debug("Found AirSensor in sensor list")
+
+                            if hasAirSensor == False:
+                                if "airTemp" in site_telem:
+                                    del site_telem["airTemp"]
+                                    _LOGGER.debug("Removed airTemp as no AirSensor was found")
+                        except KeyError as e:
+                            _LOGGER.error(f"Error processing sensors: {e}")
+                            _LOGGER.debug(f"Backyard keys: {list(config_item.get('Backyard', {}).keys())}")
+                        
+                        _LOGGER.debug(f"Adding telemetry for system {system['MspSystemID']} to results")
+                        telem_list.append(site_telem)
+                    except Exception as e:
+                        _LOGGER.error(f"Error processing system {system['MspSystemID']}: {str(e)}")
+                        _LOGGER.debug("Exception details", exc_info=True)
+            except Exception as e:
+                _LOGGER.error(f"Error getting telemetry data: {str(e)}")
+                _LOGGER.debug("Exception details", exc_info=True)
+                raise OmniLogicException(f"Failure getting telemetry: {str(e)}")
 
         else:
-            raise OmniLogicException("Failure getting telemetry.")
+            if self.token is None:
+                _LOGGER.error("Failed to get telemetry: No authentication token available")
+                raise OmniLogicException("Failure getting telemetry: No authentication token")
+            elif len(self.systems) == 0:
+                _LOGGER.error("Failed to get telemetry: No systems found")
+                raise OmniLogicException("Failure getting telemetry: No systems found")
+            else:
+                _LOGGER.error("Failed to get telemetry: Unknown reason")
+                raise OmniLogicException("Failure getting telemetry.")
 
         """
         f = open("telemetry_" + self.username + ".txt", "w")
@@ -1045,9 +1212,11 @@ class OmniLogic:
     # def get_alarm_list(self):
 
     def convert_to_json(self, xmlString):
-        my_dict = xmltodict.parse(xmlString)
-        json_data = json.dumps(my_dict)
-        # print(json_data)
+        try:
+            my_dict = xmltodict.parse(xmlString)
+            json_data = json.dumps(my_dict)
+        except:
+            raise OmniLogicException("Error converting Hayward data to JSON.")
 
         return json.loads(json_data)["Response"]["MSPConfig"]
 
